import argparse
import os
import time

import numpy as np
import so3g
import txaio
import yaml
from ocs import ocs_agent, site_config
from spt3g import core


def get_smurf_status():
    """Loads a sample status dict from file"""
    status_file = os.path.join(os.path.split(__file__)[0], 'status_sample.yaml')
    with open(status_file, 'r') as f:
        return yaml.safe_load(f)


SOSTREAM_VERSION = 2
NBIASLINES = 16
NBANDS = 8
# Range of frequencies allowed by smurf
SMURF_FREQ_RANGE = (4e3, 8e3)
SUBBANDS_PER_BAND = 512
CHANS_PER_BAND = 512

primary_names = [
    'UnixTime', 'FluxRampIncrement', 'FluxRampOffset', 'Counter0',
    'Counter1', 'Counter2', 'AveragingResetBits', 'FrameCounter',
    'TESRelaySetting'
]
primary_idxs = {name: idx for idx, name in enumerate(primary_names)}


class Tune:
    """
    Helper class for generating tunes
    """

    def __init__(self, nchans=1720):
        self.log = txaio.make_logger()

        self.nchans = nchans
        self.res_freqs = np.linspace(*SMURF_FREQ_RANGE, nchans, endpoint=False)

        band_width = (SMURF_FREQ_RANGE[1] - SMURF_FREQ_RANGE[0]) / NBANDS
        subband_width = band_width / SUBBANDS_PER_BAND

        rs = self.res_freqs - SMURF_FREQ_RANGE[0]
        self.bands = (rs / band_width).astype(int)
        self.subbands = ((rs / subband_width) % SUBBANDS_PER_BAND).astype(int)

        # just assigns channels in order for each band, making sure this
        # doesn't go above chans_per_band
        self.channels = np.full(nchans, -1, dtype=int)
        for b in np.unique(self.bands):
            m = self.bands == b
            self.channels[m] = np.arange(np.sum(m))
        self.channels[self.channels >= CHANS_PER_BAND] = -1

        self.assignment_files = [None for _ in range(NBANDS)]

    def encode_band(self, band):
        """
        Encodes band-information in the format of pysmurf tunefiles. This
        has the same structure as pysmurf tunefiles, but contains just enough
        information for indexing.
        """
        d = {
            'lock_status': {},
            'find_freq': {
                'resonance': self.res_freqs,
            },
            'tone_power': 12,
            'resonances': {}
        }

        for i, f in enumerate(self.res_freqs[self.bands == band]):
            d['resonances'][i] = {'freq': f}

        if self.assignment_files[band] is not None:
            d['channel_assignment'] = self.assignment_files[band]

        return d

    def encode_tune(self):
        """
        Encodes a full tune dictionary in the format of pysmurf tunefiles.
        """
        return {
            b: self.encode_band(b)
            for b in np.unique(self.bands)
        }

    def write_tune(self, basedir=''):
        """
        Writes tune to disk.

        Args
        ----
        basedir : str
            Directory where tune should be written.
        """
        timestamp = int(time.time())

        path = os.path.join(basedir, f'{timestamp}_tune.npy')
        np.save(path, self.encode_tune(), allow_pickle=True)
        self.tune_path = path
        self.log.debug(f"Writing tune: {self.tune_path}")
        return path

    def write_channel_assignments(self, bands=None, basedir=''):
        """
        Writes channel assignment files to disk.

        Args
        -----
        bands : optional, int, list[int]
            Bands to write to disk. Defaults to all that are present in the
            tune.
        basedir : str
            Directory where files should be written
        """
        if bands is None:
            bands = np.unique(self.bands)
        bands = np.atleast_1d(bands)

        timestamp = int(time.time())

        for b in bands:
            path = os.path.join(
                basedir, f'{timestamp}_channel_assignment_b{b}.txt'
            )
            m = self.bands == b
            d = np.array([
                self.res_freqs[m],
                self.subbands[m],
                self.channels[m],
                np.full(np.sum(m), -1)
            ]).T
            np.savetxt(path, d, fmt='%.4f,%d,%d,%d')
            self.assignment_files[b] = path


class G3FrameGenerator:
    """
    Helper class for generating G3 Streams.
    """

    def __init__(self, stream_id, sample_rate, tune,
                 action=None, action_time=None):
        self.frame_num = 0
        self.sample_num = 0
        self.session_id = int(time.time())
        self.tune = tune
        self.nchans = np.sum(tune.channels != -1)
        self.sample_rate = sample_rate
        self.stream_id = stream_id
        self.action = action
        self.action_time = action_time

    def tag_frame(self, fr):
        fr['frame_num'] = self.frame_num
        fr['session_id'] = self.session_id
        fr['sostream_id'] = self.stream_id
        fr['sostream_version'] = SOSTREAM_VERSION
        fr['time'] = core.G3Time(time.time() * core.G3Units.s)
        self.frame_num += 1
        return fr

    def get_obs_start_frame(self):
        fr = core.G3Frame(core.G3FrameType.Observation)
        fr['stream_placement'] = 'start'
        self.tag_frame(fr)
        return fr

    def get_obs_end_frame(self):
        fr = core.G3Frame(core.G3FrameType.Observation)
        fr['stream_placement'] = 'end'
        self.tag_frame(fr)
        return fr

    def get_status_frame(self):
        fr = core.G3Frame(core.G3FrameType.Wiring)
        s = get_smurf_status()

        tune_key = 'AMCc.FpgaTopLevel.AppTop.AppCore.SysgenCryo.tuneFilePath'
        s[tune_key] = self.tune.tune_path

        m = self.tune.channels != -1
        chmask = self.tune.channels[m] + self.tune.bands[m] * CHANS_PER_BAND

        s['AMCc.SmurfProcessor.ChannelMapper.Mask'] = str(chmask.tolist())
        s['AMCc.SmurfProcessor.ChannelMapper.NumChannels'] = self.nchans.item()

        pysmurf_root = "AMCc.SmurfProcessor.SOStream"
        if self.action is not None:
            s[f'{pysmurf_root}.pysmurf_action'] = self.action
        if self.action_time is not None:
            s[f'{pysmurf_root}.pysmurf_action_timestamp'] = int(self.action_time)

        fr['status'] = yaml.dump(s)
        fr['dump'] = True
        self.tag_frame(fr)
        return fr

    def get_data_frame(self, start, stop):

        times = np.arange(start, stop, 1. / self.sample_rate)
        nsamps = len(times)
        frame_counter = np.arange(self.sample_num, self.sample_num + nsamps, dtype=int)
        self.sample_num += nsamps
        chans = np.arange(self.nchans)
        names = [f'r{ch:0>4}' for ch in chans]

        count_per_phi0 = 2**16
        data = np.zeros((self.nchans, nsamps), dtype=np.int32)
        data += count_per_phi0 * chans[:, None]
        data += (count_per_phi0 * 0.2 * np.sin(2 * np.pi * 8 * times)).astype(int)
        data += (count_per_phi0 * np.random.normal(0, 0.03, (self.nchans, nsamps))).astype(int)

        fr = core.G3Frame(core.G3FrameType.Scan)

        g3times = core.G3VectorTime(times * core.G3Units.s)
        fr['data'] = so3g.G3SuperTimestream(names, g3times, data)

        primary_data = np.zeros((len(primary_names), nsamps), dtype=np.int64)
        primary_data[primary_idxs['UnixTime'], :] = (times * 1e9).astype(int)
        primary_data[primary_idxs['FrameCounter'], :] = frame_counter
        fr['primary'] = so3g.G3SuperTimestream(primary_names, g3times, primary_data)

        tes_bias_names = [f'bias{bg:0>2}' for bg in range(NBIASLINES)]
        bias_data = np.zeros((NBIASLINES, nsamps), dtype=np.int32)
        fr['tes_biases'] = so3g.G3SuperTimestream(tes_bias_names, g3times, bias_data)

        fr['timing_paradigm'] = 'Low Precision'
        fr['num_samples'] = nsamps

        self.tag_frame(fr)
        return fr


class DataStreamer:
    """
    Helper class for streaming G3 data
    """

    def __init__(self, stream_id, sample_rate, tune, timestreamdir,
                 file_duration, frame_len, action=None, action_time=None):
        self.frame_gen = G3FrameGenerator(stream_id, sample_rate, tune,
                                          action=action, action_time=action_time)
        self.session_id = self.frame_gen.session_id
        self.stream_id = stream_id
        self.timestreamdir = timestreamdir
        self.seq = 0
        self.file_duration = file_duration
        self.file_start = 0
        self.writer = None
        self.file_list = []
        self.frame_len = frame_len

    def _get_g3_filename(self):
        """
        Returns the file path for a g3-file with specified session id and seq
        idx.
        """
        timecode = f"{self.session_id}"[:5]
        subdir = os.path.join(self.timestreamdir, timecode, self.stream_id)
        filepath = os.path.join(subdir, f"{self.session_id}_{self.seq:0>3}.g3")
        return filepath

    def _new_file(self):
        """
        Ends the current G3File (if one is open) and begins a new one,
        incrementing ``seq`` after updating.
        """
        self.end_file()
        fname = self._get_g3_filename()
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        self.writer = core.G3Writer(fname)
        if self.seq == 0:
            self.writer(self.frame_gen.get_obs_start_frame())
            self.writer(self.frame_gen.get_status_frame())
        self.file_start = time.time()
        self.file_list.append(fname)
        self.seq += 1

    def end_file(self):
        """
        Ends the current file by sending a G3EndProcessing Frame.
        """
        if self.writer is not None:
            self.writer(self.frame_gen.get_obs_end_frame())
            self.writer(core.G3Frame(core.G3FrameType.EndProcessing))

    def write_next(self):
        """
        Writes the next data frame to disk. Will rotate files based on the
        current file start time and the file duration. This sleep wait
        for the frame-duration before writing the G3Frame to disk.
        """
        start = time.time()
        if (start - self.file_start > self.file_duration) or (self.writer is None):
            self._new_file()
        time.sleep(self.frame_len)
        stop = time.time()
        self.writer(self.frame_gen.get_data_frame(start, stop))

    def stream_between(self, start, stop, wait=False):
        """
        This function will create a new observation and "stream" data between
        a specified start and stop time. This function will by default generate
        and write the data without sleeping for the specified amount of time.
        To avoid confusion, this will not rotate G3Files since that gets kind
        of complicated when you're not running in real time.

        Args
        ------
        start : float
            Start time of data
        stop : float
            Stop time of data
        wait : bool
            If True, will sleep for the correct amount of time between each
            written frame. Defaults to False.
        """
        frame_starts = np.arange(start, stop, self.frame_len)
        frame_stops = frame_starts + self.frame_len

        # In case there's already an open file
        self.seq = 0
        self.end_file()

        self._new_file()
        for t0, t1 in zip(frame_starts, frame_stops):
            if wait:
                now = time.time()
                if now < t1:
                    time.sleep(t1 - now)
            self.writer(self.frame_gen.get_data_frame(start, stop))
        self.end_file()


class SmurfFileEmulator:
    """
    OCS Agent for emulating file creation for the smurf system.
    """

    def __init__(self, agent, args):
        self.log = agent.log
        self.file_duration = args.file_duration
        self.stream_id = args.stream_id
        self.basedir = args.base_dir
        if not os.path.exists(self.basedir):
            raise ValueError(f"Basedir {self.basedir} does not exist")

        self.smurfdir = os.path.join(self.basedir, 'smurf')
        self.timestreamdir = os.path.join(self.basedir, 'timestreams')
        self.nchans = args.nchans
        self.sample_rate = args.sample_rate
        self.frame_len = args.frame_len

        self.streaming = False
        self.tune = None

    def _new_streamer(self, action=None, action_time=None):
        return DataStreamer(
            self.stream_id, self.sample_rate, self.tune, self.timestreamdir,
            self.file_duration, self.frame_len,
            action=action, action_time=action_time,
        )

    def _get_action_dir(self, action, action_time=None, is_plot=False):
        t = int(time.time())
        if action_time is None:
            action_time = t
        action_time = int(action_time)
        timecode = f"{action_time}"[:5]
        dir_type = 'plots' if is_plot else 'outputs'
        subdir = os.path.join(
            self.smurfdir, timecode, self.stream_id, f'{action_time}_{action}',
            dir_type
        )
        os.makedirs(subdir, exist_ok=True)
        return subdir

    def _write_smurf_file(self, name, action, action_time=None,
                          prepend_ctime=True, is_plot=False):
        """
        Creates a fake pysmurf ancilliary file containing just the creation
        time of the file.

        For example::

            self._write_smurf_file('IV.npy', 'take_IV')

        will create the file::

            data/smurf/16371/<stream_id>/1637182065_take_IV/1637182065_IV.npy

        Args:
            name (str):
                Name of the file to be created. This should not include the
                ctime since the ctime will be prepended. I.e. "IV.npy" will
                become "<ctime>_IV.npy"
            action (str):
                Pysmurf action name for determining the subdirectory this
                should be written to.
            action_time (float):
                Action timestamp for the file. If None, will set to the current
                time.
        """
        t = int(time.time())
        if action_time is None:
            action_time = t
        subdir = self._get_action_dir(
            action, action_time=action_time, is_plot=is_plot
        )
        if prepend_ctime:
            filepath = os.path.join(subdir, f'{t}_{name}')
        else:
            filepath = os.path.join(subdir, name)
        self.log.info(f"Writing smurf file: {filepath}")
        with open(filepath, 'w') as f:
            f.write(f'start: {time.time()}\n')
        return filepath

    def _run_setup(self, action=None, action_time=None, session=None):
        """
        Helper function to run setup operations, used by uxm_setup and
        uxm_relock.
        """
        self.tune = Tune(nchans=self.nchans)

        if action is None:
            action = 'uxm_setup'
        if action_time is None:
            action_time = time.time()

        # Find Freq
        files = ['amp_sweep_freq.txt', 'amp_sweep_resonance.txt',
                 'amp_sweep_resp.txt']
        for f in files:
            self._write_smurf_file(f, action, action_time=action_time)

        # Setup Notches
        sdir = self._get_action_dir(action, action_time=action_time)
        self.tune.write_channel_assignments(basedir=sdir)
        self.tune.write_tune(basedir=sdir)

        # tracking setup
        action_time = time.time()
        fname = f"{int(time.time())}.dat"
        self._write_smurf_file(fname, action, action_time=action_time, prepend_ctime=False)

        # Short g3 stream
        streamer = self._new_streamer(action=action, action_time=action_time)
        now = time.time()
        streamer.stream_between(now, now + 30, wait=False)
        if session is not None:
            session.data['noise_file'] = streamer.file_list[0]

    def uxm_setup(self, session, params):
        """uxm_setup(test_mode=False)

        **Task** - Emulates files that might come from a general tune dets
        function. These are some of the files found on simons1 registered when
        running the following ops with a single band:

             1. Find-freq
             2. setup_notches
             3. tracking_setup
             4. short g3 stream
        """
        self._run_setup(action='uxm_setup', action_time=time.time(),
                        session=session)
        time.sleep(1)
        return True, "Wrote tune files"

    def take_noise(self, session, params=None):
        """take_noise()

        **Task** - Takes a short noise timestream
        """
        action = 'take_noise'
        action_time = time.time()

        streamer = self._new_streamer(action=action, action_time=action_time)
        now = time.time()
        streamer.stream_between(now, now + 30, wait=False)
        session.data['noise_file'] = streamer.file_list[0]
        time.sleep(1)
        return True, "Took noise data"

    def uxm_relock(self, session, params=None):
        """uxm_relock()

        **Task** - Normally this wouldn't involve a full find-freq, but for
        emulation purposes it's ok if this is the same as uxm_setup.
        """
        self._run_setup(action='uxm_relock', action_time=time.time(),
                        session=session)
        time.sleep(1)
        return True, "Wrote tune files"

    @ocs_agent.param('wait', default=True)
    def take_iv(self, session, params=None):
        """take_iv()

        **Task** - Creates files generated associated with iv taking / analysis
        """
        action = 'take_iv'
        action_time = time.time()
        files = ['iv_analyze.npy', 'iv_bias_all.npy', 'iv_info.npy']

        streamer = self._new_streamer(action=action, action_time=action_time)
        now = time.time()
        streamer.stream_between(now, now + 5, wait=params['wait'])

        for f in files:
            self._write_smurf_file(f, action, action_time=action_time)

        return True, "Wrote IV files"

    @ocs_agent.param('wait', default=True)
    def take_bias_steps(self, session, params=None):
        """take_bias_steps()

        **Task** - Creates files associated with taking bias steps
        """
        action = 'take_bias_steps'
        action_time = time.time()
        files = ['bias_step_analysis.npy']

        streamer = self._new_streamer(action=action, action_time=action_time)
        now = time.time()
        streamer.stream_between(now, now + 5, wait=params['wait'])

        for f in files:
            self._write_smurf_file(f, action, action_time=action_time)

        return True, "Wrote Bias Step Files"

    @ocs_agent.param('wait', default=True)
    def take_bgmap(self, session, params=None):
        """take_bgmap()

        **Task** - Creates files associated with taking a bias group mapping.
        """
        action = 'take_bgmap'
        action_time = time.time()

        streamer = self._new_streamer(action=action, action_time=action_time)
        now = time.time()
        streamer.stream_between(now, now + 5, wait=params['wait'])

        files = ['bg_map.npy', 'bias_step_analysis.npy']
        for f in files:
            self._write_smurf_file(f, action, action_time=action_time)

        return True, "Finished taking bgmap"

    def bias_dets(self, session, params=None):
        """bias_dets()

        **Task** - Creates files associated with biasing dets, which is none.
        """
        time.sleep(1)
        return True, 'Wrote det biasing files'

    @ocs_agent.param('duration', default=None)
    def stream(self, session, params):
        """stream(duration=None)

        **Process** - Generates example fake-files organized in the same way as
        they would be a regular smurf-stream. For end-to-end testing, we want
        an example of a pysmurf-ancilliary file, and then regular g3 that rotate
        at regular intervals. The content of the files here don't match what
        actual G3 or pysmurf files look like, however the directory structure
        is the same.

        Parameters:
            duration (float, optional):
                If set, will stop stream after specified amount of time (sec).
        """
        session.set_status('starting')

        if self.tune is None:
            raise ValueError("No tune loaded!")

        # Write initial smurf metadata
        if 'duration' in params:
            action = 'take_g3_data'
        else:
            action = 'stream_g3_on'

        action_time = time.time()
        files = ['freq.txt', 'mask.txt']
        for f in files:
            self._write_smurf_file(f, action, action_time=action_time)

        end_time = None
        if params.get('duration') is not None:
            end_time = time.time() + params['duration']

        session.set_status('running')
        streamer = self._new_streamer(action=action, action_time=action_time)
        session.data['session_id'] = streamer.session_id
        session.data['g3_files'] = streamer.file_list

        self.streaming = True
        while self.streaming:
            streamer.write_next()

            if end_time is not None:
                if time.time() > end_time:
                    break

        streamer.end_file()

        time.sleep(1)
        return True, "Finished Stream"

    def _stop_stream(self, session, params=None):
        if self.streaming:
            session.set_status('stopping')
            self.streaming = False
            return True, 'requesting to stop taking data'
        else:
            return False, 'agent is not currently streaming'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--stream-id', required=True,
                        help='Stream ID for fake smurf stream')
    pgroup.add_argument('--base-dir', required=True,
                        help="Base directory where data should be written")
    pgroup.add_argument('--file-duration', default=10 * 60, type=float,
                        help="Time in sec before rotating g3 files")
    pgroup.add_argument('--nchans', default=1024, type=int,
                        help="Number of channels to stream from")
    pgroup.add_argument('--sample-rate', default=200, type=float,
                        help="Sample rate for streaming data")
    pgroup.add_argument('--frame-len', default=2, type=float,
                        help="Time per G3 data frame (seconds)")

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='SmurfFileEmulator',
                                  parser=parser,
                                  args=args)

    txaio.start_logging(level=os.environ.get('LOGLEVEL', 'info'))

    agent, runner = ocs_agent.init_site_agent(args)

    file_em = SmurfFileEmulator(agent, args)
    agent.register_task('uxm_setup', file_em.uxm_setup)
    agent.register_task('uxm_relock', file_em.uxm_relock)
    agent.register_task('take_iv', file_em.take_iv)
    agent.register_task('take_bias_steps', file_em.take_bias_steps)
    agent.register_task('take_bgmap', file_em.take_bgmap)
    agent.register_task('bias_dets', file_em.bias_dets)
    agent.register_task('take_noise', file_em.take_noise)
    agent.register_process('stream', file_em.stream, file_em._stop_stream)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import calendar
import datetime
import struct
import time

import soaculib as aculib
import soaculib.status_keys as status_keys
import twisted.web.client as tclient
from autobahn.twisted.util import sleep as dsleep
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from soaculib.twisted_backend import TwistedHttpBackend
from twisted.internet import protocol, reactor
from twisted.internet.defer import inlineCallbacks

import socs.agents.acu.drivers as sh


def timecode(acutime):
    """
    Takes the time code produced by the ACU status stream and returns
    a ctime.

    Parameters:
        acutime (float): The time recorded by the ACU status stream,
                         corresponding to the fractional day of the year
    """
    sec_of_day = (acutime - 1) * 60 * 60 * 24
    year = datetime.datetime.now().year
    gyear = calendar.timegm(time.strptime(str(year), '%Y'))
    comptime = gyear + sec_of_day
    return comptime


def uploadtime_to_ctime(ptstack_time, upload_year):
    year = int(upload_year)
    gyear = calendar.timegm(time.strptime(str(year), '%Y'))
    day_of_year = float(ptstack_time.split(',')[0]) - 1.0
    hour = float(ptstack_time.split(',')[1].split(':')[0])
    minute = float(ptstack_time.split(',')[1].split(':')[1])
    second = float(ptstack_time.split(',')[1].split(':')[2])
    comptime = gyear + day_of_year * 60 * 60 * 24 + hour * 60 * 60 + minute * 60 + second
    return comptime


def pop_first_vals(data_dict, group_size):
    new_data_dict = {}
    for key in data_dict.keys():
        if len(data_dict[key]):
            new_data_dict[key] = data_dict[key][group_size:]
        else:
            print('no more data')
    return new_data_dict


def front_group(data_dict, group_size):
    new_data_dict = {}
    for key in data_dict.keys():
        if len(data_dict[key]) > group_size:
            new_data_dict[key] = data_dict[key][:group_size]
        else:
            new_data_dict[key] = data_dict[key]
    return new_data_dict


class ACUAgent:
    """
    Agent to acquire data from an ACU and control telescope pointing with the
    ACU.

    Parameters:
        acu_config (str):
            The configuration for the ACU, as referenced in aculib.configs.
            Default value is 'guess'.

    """

    def __init__(self, agent, acu_config='guess'):
        self.lock = TimeoutLock()
        self.jobs = {
            'monitor': 'idle',
            'broadcast': 'idle',
            'control': 'idle',  # shared by all motion tasks/processes
            'scanspec': 'idle',
        }

        self.acu_config = aculib.guess_config(acu_config)
        self.sleeptime = self.acu_config['motion_waittime']
        self.udp = self.acu_config['streams']['main']
        self.udp_schema = aculib.get_stream_schema(self.udp['schema'])
        self.udp_ext = self.acu_config['streams']['ext']
        self.acu8100 = self.acu_config['status']['status_name']
        self.monitor_fields = status_keys.status_fields[self.acu_config['platform']]['status_fields']
        self.motion_limits = self.acu_config['motion_limits']

        self.log = agent.log

        # self.data provides a place to reference data from the monitors.
        # 'status' is populated by the monitor operation
        # 'broadcast' is populated by the udp_monitor operation

        self.data = {'status': {'summary': {},
                                'position_errors': {},
                                'axis_limits': {},
                                'axis_faults_errors_overages': {},
                                'axis_warnings': {},
                                'axis_failures': {},
                                'axis_state': {},
                                'osc_alarms': {},
                                'commands': {},
                                'ACU_failures_errors': {},
                                'platform_status': {},
                                'ACU_emergency': {},
                                },
                     'broadcast': {},
                     'uploads': {'Start_Azimuth': 0.0,
                                 'Start_Elevation': 0.0,
                                 'Start_Boresight': 0.0,
                                 'Command_Type': 0,
                                 'Preset_Azimuth': 0.0,
                                 'Preset_Elevation': 0.0,
                                 'Preset_Boresight': 0.0,
                                 'PtStack_Lines': 'False',
                                 'PtStack_Time': '000, 00:00:00.000000',
                                 'PtStack_Azimuth': 0.0,
                                 'PtStack_Elevation': 0.0,
                                 'PtStack_AzVelocity': 0.0,
                                 'PtStack_ElVelocity': 0.0,
                                 'PtStack_AzFlag': 0,
                                 'PtStack_ElFlag': 0},
                     'scanspec': {},
                     }

        self.agent = agent

        self.take_data = False

        # self.web_agent = tclient.Agent(reactor)
        tclient._HTTP11ClientFactory.noisy = False

        self.acu_control = aculib.AcuControl(
            acu_config, backend=TwistedHttpBackend(persistent=False))
        self.acu_read = aculib.AcuControl(
            acu_config, backend=TwistedHttpBackend(persistent=True), readonly=True)

        agent.register_process('monitor',
                               self.monitor,
                               lambda session, params: self._set_job_stop('monitor'),
                               blocking=False,
                               startup=True)
        agent.register_process('broadcast',
                               self.broadcast,
                               lambda session, params: self._set_job_stop('broadcast'),
                               blocking=False,
                               startup=True)
        agent.register_process('generate_scan',
                               self.generate_scan,
                               lambda session, params: self._set_job_stop('control'),
                               blocking=False,
                               startup=False)
        basic_agg_params = {'frame_length': 60}
        fullstatus_agg_params = {'frame_length': 60,
                                 'exclude_influx': True,
                                 'exclude_aggregator': False
                                 }
        influx_agg_params = {'frame_length': 60,
                             'exclude_influx': False,
                             'exclude_aggregator': True
                             }
        self.stop_times = {'az': float('nan'),
                           'el': float('nan'),
                           'bs': float('nan'),
                           }
        self.agent.register_feed('acu_status',
                                 record=True,
                                 agg_params=fullstatus_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_status_influx',
                                 record=True,
                                 agg_params=influx_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_commands_influx',
                                 record=True,
                                 agg_params=influx_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_udp_stream',
                                 record=True,
                                 agg_params=fullstatus_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_broadcast_influx',
                                 record=True,
                                 agg_params=influx_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_upload',
                                 record=True,
                                 agg_params=basic_agg_params,
                                 buffer_time=1)
        self.agent.register_feed('acu_error',
                                 record=True,
                                 agg_params=basic_agg_params,
                                 buffer_time=1)
        agent.register_task('go_to', self.go_to, blocking=False)
        agent.register_task('constant_velocity_scan',
                            self.constant_velocity_scan,
                            blocking=False)
        agent.register_task('fromfile_scan',
                            self.fromfile_scan,
                            blocking=False)
        agent.register_task('set_boresight',
                            self.set_boresight,
                            blocking=False)
        agent.register_task('stop_and_clear',
                            self.stop_and_clear,
                            blocking=False)

    # Operation management.  This agent has several Processes that
    # must be able to alone or simultaneously.  The state of each is
    # registered in self.jobs, protected by self.lock (though this is
    # probably not necessary as long as we don't thread).  Any logic
    # to assess conflicts should probably be in _try_set_job.

    def _try_set_job(self, job_name):
        """
        Set a job status to 'run'.

        Parameters:
            job_name (str): Name of the task/process you are trying to start.
        """
        with self.lock.acquire_timeout(timeout=1.0, job=job_name) as acquired:
            if not acquired:
                self.log.warn("Lock could not be acquired because it is held"
                              f" by {self.lock.job}")
                return False
            # Set running.
            self.jobs[job_name] = 'run'
            return (True, 'ok')

    def _set_job_stop(self, job_name):
        """
        Set a job status to 'stop'.

        Parameters:
            job_name (str): Name of the process you are trying to stop.
        """
        print('try to acquire stop')
        # return (False, 'Could not stop')
        with self.lock.acquire_timeout(timeout=1.0, job=job_name) as acquired:
            if not acquired:
                self.log.warn("Lock could not be acquired because it is"
                              f" held by {self.lock.job}")
                return False
            try:
                self.jobs[job_name] = 'stop'
            # state = self.jobs.get(job_name, 'idle')
            # if state == 'idle':
            #     return False, 'Job not running.'
            # if state == 'stop':
            #     return False, 'Stop already requested.'
            # self.jobs[job_name] = 'stop'
                return True, 'Requested Process stop.'
            except Exception as e:
                print(str(e))

    def _set_job_done(self, job_name):
        """
        Set a job status to 'idle'.

        Parameters:
            job_name (str): Name of the task/process you are trying to idle.
        """
        with self.lock.acquire_timeout(timeout=1.0, job=job_name) as acquired:
            if not acquired:
                self.log.warn("Lock could not be acquried because it is held"
                              f" by {self.lock.job}")
                return False
            self.jobs[job_name] = 'idle'

    #
    # The Operations
    #

    @inlineCallbacks
    def monitor(self, session, params):
        """monitor()

        **Process** - Refresh the cache of SATP ACU status information and
        report it on the 'acu_status_summary' and 'acu_status_full' HK feeds.

        Summary parameters are ACU-provided time code, Azimuth mode,
        Azimuth position, Azimuth velocity, Elevation mode, Elevation position,
        Elevation velocity, Boresight mode, and Boresight position.

        """
        ok, msg = self._try_set_job('monitor')
        if not ok:
            return ok, msg

        session.set_status('running')
        version = yield self.acu_read.http.Version()
        self.log.info(version)

        mode_key = {'Stop': 0,
                    'Preset': 1,
                    'ProgramTrack': 2,
                    'Stow': 3,
                    'SurvivalMode': 4,
                    'Rate': 5,
                    'StarTrack': 6,
                    }
        tfn_key = {'None': float('nan'),
                   'False': 0,
                   'True': 1,
                   }
        report_t = time.time()
        report_period = 20
        n_ok = 0
        min_query_period = 0.05   # Seconds
        query_t = 0
        while self.jobs['monitor'] == 'run':
            now = time.time()

            if now > report_t + report_period:
                self.log.info('Responses ok at %.3f Hz'
                              % (n_ok / (now - report_t)))
                report_t = now
                n_ok = 0

            if now - query_t < min_query_period:
                yield dsleep(min_query_period - (now - query_t))

            query_t = time.time()
            try:
                j = yield self.acu_read.http.Values(self.acu8100)
                n_ok += 1
                session.data = j
            except Exception as e:
                # Need more error handling here...
                errormsg = {'aculib_error_message': str(e)}
                self.log.error(str(e))
                acu_error = {'timestamp': time.time(),
                             'block_name': 'ACU_error',
                             'data': errormsg
                             }
                self.agent.publish_to_feed('acu_error', acu_error)
                yield dsleep(1)
                continue

            for (key, value) in session.data.items():
                for category in self.monitor_fields:
                    if key in self.monitor_fields[category]:
                        if isinstance(value, bool):
                            self.data['status'][category][self.monitor_fields[category][key]] = int(value)
                        elif isinstance(value, int) or isinstance(value, float):
                            self.data['status'][category][self.monitor_fields[category][key]] = value
                        elif value is None:
                            self.data['status'][category][self.monitor_fields[category][key]] = float('nan')
                        else:
                            self.data['status'][category][self.monitor_fields[category][key]] = str(value)
            self.data['status']['summary']['ctime'] =\
                timecode(self.data['status']['summary']['Time'])

            # influx_status refers to all other self.data['status'] keys. Do not add
            # more keys to any self.data['status'] categories beyond this point
            influx_status = {}
            for category in self.data['status']:
                if category != 'commands':
                    for statkey, statval in self.data['status'][category].items():
                        if isinstance(statval, float):
                            influx_status[statkey + '_influx'] = statval
                        elif isinstance(statval, str):
                            if statval == 'None':
                                influx_status[statkey + '_influx'] = float('nan')
                            elif statval in ['True', 'False']:
                                influx_status[statkey + '_influx'] = tfn_key[statval]
                            else:
                                influx_status[statkey + '_influx'] = mode_key[statval]
                        elif isinstance(statval, int):
                            if statkey in ['Year', 'Free_upload_positions']:
                                influx_status[statkey + '_influx'] = float(statval)
                            else:
                                influx_status[statkey + '_influx'] = int(statval)
                elif category == 'commands':
                    if str(self.data['status']['commands']['Azimuth_commanded_position']) != 'nan':
                        acucommand_az = {'timestamp': self.data['status']['summary']['ctime'],
                                         'block_name': 'ACU_commanded_positions_az',
                                         'data': {'Azimuth_commanded_position_influx': self.data['status']['commands']['Azimuth_commanded_position']}
                                         }
                        self.agent.publish_to_feed('acu_commands_influx', acucommand_az)
                    if str(self.data['status']['commands']['Elevation_commanded_position']) != 'nan':
                        acucommand_el = {'timestamp': self.data['status']['summary']['ctime'],
                                         'block_name': 'ACU_commanded_positions_el',
                                         'data': {'Elevation_commanded_position_influx': self.data['status']['commands']['Elevation_commanded_position']}
                                         }
                        self.agent.publish_to_feed('acu_commands_influx', acucommand_el)
                    if str(self.data['status']['commands']['Boresight_commanded_position']) != 'nan':
                        acucommand_bs = {'timestamp': self.data['status']['summary']['ctime'],
                                         'block_name': 'ACU_commanded_positions_boresight',
                                         'data': {'Boresight_commanded_position_influx': self.data['status']['commands']['Boresight_commanded_position']}
                                         }
                        self.agent.publish_to_feed('acu_commands_influx', acucommand_bs)
            if self.data['uploads']['PtStack_Time'] == '000, 00:00:00.000000':
                self.data['uploads']['PtStack_ctime'] = self.data['status']['summary']['ctime']

            acustatus_summary = {'timestamp':
                                 self.data['status']['summary']['ctime'],
                                 'block_name': 'ACU_summary_output',
                                 'data': self.data['status']['summary']
                                 }
            acustatus_axisfaults = {'timestamp': self.data['status']['summary']['ctime'],
                                    'block_name': 'ACU_axis_faults',
                                    'data': self.data['status']['axis_faults_errors_overages']
                                    }
            acustatus_poserrors = {'timestamp': self.data['status']['summary']['ctime'],
                                   'block_name': 'ACU_position_errors',
                                   'data': self.data['status']['position_errors']
                                   }
            acustatus_axislims = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_axis_limits',
                                  'data': self.data['status']['axis_limits']
                                  }
            acustatus_axiswarn = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_axis_warnings',
                                  'data': self.data['status']['axis_warnings']
                                  }
            acustatus_axisfail = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_axis_failures',
                                  'data': self.data['status']['axis_failures']
                                  }
            acustatus_axisstate = {'timestamp': self.data['status']['summary']['ctime'],
                                   'block_name': 'ACU_axis_state',
                                   'data': self.data['status']['axis_state']
                                   }
            acustatus_oscalarm = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_oscillation_alarm',
                                  'data': self.data['status']['osc_alarms']
                                  }
            acustatus_commands = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_command_status',
                                  'data': self.data['status']['commands']
                                  }
            acustatus_acufails = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_general_errors',
                                  'data': self.data['status']['ACU_failures_errors']
                                  }
            acustatus_platform = {'timestamp': self.data['status']['summary']['ctime'],
                                  'block_name': 'ACU_platform_status',
                                  'data': self.data['status']['platform_status']
                                  }
            acustatus_emergency = {'timestamp': self.data['status']['summary']['ctime'],
                                   'block_name': 'ACU_emergency',
                                   'data': self.data['status']['ACU_emergency']
                                   }
            acustatus_influx = {'timestamp':
                                self.data['status']['summary']['ctime'],
                                'block_name': 'ACU_status_INFLUX',
                                'data': influx_status
                                }
            status_blocks = [acustatus_summary, acustatus_axisfaults, acustatus_poserrors,
                             acustatus_axislims, acustatus_axiswarn, acustatus_axisfail,
                             acustatus_axisstate, acustatus_oscalarm, acustatus_commands,
                             acustatus_acufails, acustatus_platform, acustatus_emergency]
            for block in status_blocks:
                self.agent.publish_to_feed('acu_status', block)
            self.agent.publish_to_feed('acu_status_influx', acustatus_influx, from_reactor=True)

        # self._set_job_stop('monitor')
        # yield dsleep(1)
        # self._set_job_done('monitor')
        return True, 'Acquisition exited cleanly.'

    @inlineCallbacks
    def broadcast(self, session, params):
        """broadcast()

        **Process** - Read UDP data from the port specified by self.acu_config,
        decode it, and publish to an HK feed.

        """
        ok, msg = self._try_set_job('broadcast')
        if not ok:
            return ok, msg
        session.set_status('running')
        FMT = self.udp_schema['format']
        FMT_LEN = struct.calcsize(FMT)
        UDP_PORT = self.udp['port']
        udp_data = []
        fields = self.udp_schema['fields']

        class MonitorUDP(protocol.DatagramProtocol):

            def datagramReceived(self, data, src_addr):
                host, port = src_addr
                offset = 0
                while len(data) - offset >= FMT_LEN:
                    d = struct.unpack(FMT, data[offset:offset + FMT_LEN])
                    udp_data.append(d)
                    offset += FMT_LEN
        handler = reactor.listenUDP(int(UDP_PORT), MonitorUDP())
        while self.jobs['broadcast'] == 'run':
            if udp_data:
                process_data = udp_data[:200]
                udp_data = udp_data[200:]
                year = datetime.datetime.now().year
                gyear = calendar.timegm(time.strptime(str(year), '%Y'))
                if len(process_data):
                    sample_rate = (len(process_data)
                                   / ((process_data[-1][0] - process_data[0][0]) * 86400
                                   + process_data[-1][1] - process_data[0][1]))
                else:
                    sample_rate = 0.0
                latest_az = process_data[2]
                latest_el = process_data[3]
                latest_az_raw = process_data[4]
                latest_el_raw = process_data[5]
                session.data = {'sample_rate': sample_rate,
                                'latest_az': latest_az,
                                'latest_el': latest_el,
                                'latest_az_raw': latest_az_raw,
                                'latest_el_raw': latest_el_raw
                                }
                bcast_first = {}
                pd0 = process_data[0]
                pd0_gday = (pd0[0] - 1) * 86400
                pd0_sec = pd0[1]
                pd0_data_ctime = gyear + pd0_gday + pd0_sec
                bcast_first['Time_bcast_influx'] = pd0_data_ctime
                for i in range(2, len(pd0)):
                    bcast_first[fields[i].replace(' ', '_') + '_bcast_influx'] = pd0[i]
                acu_broadcast_influx = {'timestamp': bcast_first['Time_bcast_influx'],
                                        'block_name': 'ACU_position_bcast_influx',
                                        'data': bcast_first,
                                        }
                self.agent.publish_to_feed('acu_broadcast_influx', acu_broadcast_influx, from_reactor=True)
                for d in process_data:
                    gday = (d[0] - 1) * 86400
                    sec = d[1]
                    data_ctime = gyear + gday + sec
                    self.data['broadcast']['Time'] = data_ctime
                    for i in range(2, len(d)):
                        self.data['broadcast'][fields[i].replace(' ', '_')] = d[i]
                    acu_udp_stream = {'timestamp': self.data['broadcast']['Time'],
                                      'block_name': 'ACU_broadcast',
                                      'data': self.data['broadcast']
                                      }
                    # print(acu_udp_stream)
                    self.agent.publish_to_feed('acu_udp_stream',
                                               acu_udp_stream, from_reactor=True)
            else:
                yield dsleep(1)
            yield dsleep(0.005)

        handler.stopListening()
        self._set_job_done('broadcast')
        return True, 'Acquisition exited cleanly.'

    @ocs_agent.param('az', type=float)
    @ocs_agent.param('el', type=float)
    @ocs_agent.param('wait', default=1., type=float)
    @ocs_agent.param('end_stop', default=False, type=bool)
    @ocs_agent.param('rounding', default=1, type=int)
    @inlineCallbacks
    def go_to(self, session, params):
        """go_to(az=None, el=None, wait=1., end_stop=False, rounding=1)

        **Task** - Move the telescope to a particular point (azimuth,
        elevation) in Preset mode. When motion has ended and the telescope
        reaches the preset point, it returns to Stop mode and ends.

        Parameters:
            az (float): destination angle for the azimuthal axis
            el (float): destination angle for the elevation axis
            wait (float): amount of time to wait for motion to end
            end_stop (bool): put the telescope in Stop mode at the end of
                the motion
            rounding (int): number of decimal places to round to

        """
        ok, msg = self._try_set_job('control')
        if not ok:
            return ok, msg
        az = params['az']
        el = params['el']
        if az <= self.motion_limits['azimuth']['lower'] or az >= self.motion_limits['azimuth']['upper']:
            raise ocs_agent.ParamError("Azimuth out of range! Must be "
                                       + f"{self.motion_limits['azimuth']['lower']} < az "
                                       + f"< {self.motion_limits['azimuth']['upper']}")
        if el <= self.motion_limits['elevation']['lower'] or el >= self.motion_limits['elevation']['upper']:
            raise ocs_agent.ParamError("Elevation out of range! Must be "
                                       + f"{self.motion_limits['elevation']['lower']} < el "
                                       + f"< {self.motion_limits['elevation']['upper']}")
        end_stop = params['end_stop']
        wait_for_motion = params['wait']
        round_int = params['rounding']
        self.log.info('Azimuth commanded position: ' + str(az))
        self.log.info('Elevation commanded position: ' + str(el))
        current_az = round(self.data['broadcast']['Corrected_Azimuth'], 4)
        current_el = round(self.data['broadcast']['Corrected_Elevation'], 4)
        self.data['uploads']['Start_Azimuth'] = current_az
        self.data['uploads']['Start_Elevation'] = current_el
        self.data['uploads']['Command_Type'] = 1
        self.data['uploads']['Preset_Azimuth'] = az
        self.data['uploads']['Preset_Elevation'] = el

        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload)

        # Check whether the telescope is already at the point
        # self.log.info('Checking current position')
        yield self.acu_control.mode('Preset')
        if round(current_az, round_int) == az and \
                round(current_el, round_int) == el:
            yield self.acu_control.go_to(az, el, wait=0.1)
            self.log.info('Already at commanded position.')
            self._set_job_done('control')
            return True, 'Preset at commanded position'
        # yield self.acu.stop()
        # yield self.acu_control.mode('Stop')
        # self.log.info('Stopped')
        # yield dsleep(0.5)
        yield self.acu_control.go_to(az, el, wait=0.1)
        mdata = self.data['status']['summary']
        # Wait for telescope to start moving
        self.log.info('Moving to commanded position')
        wait_for_motion_start = time.time()
        elapsed_wait_for_motion = 0.0
        while mdata['Azimuth_current_velocity'] == 0.0 and\
                mdata['Elevation_current_velocity'] == 0.0:
            if elapsed_wait_for_motion < 30.:
                yield dsleep(wait_for_motion)
                elapsed_wait_for_motion = time.time() - wait_for_motion_start
                mdata = self.data['status']['summary']
            else:
                if round(mdata['Azimuth_current_position'] - az, round_int) == 0. and \
                        round(mdata['Elevation_current_position'] - el, round_int) == 0.:
                    if end_stop:
                        yield self.acu_control.stop()
                        self.log.info('Az and el in Stop mode')
                    self._set_job_done('control')
                    return True, 'Pointing completed'
                else:
                    yield self.acu_control.stop()
                    self._set_job_done('control')
                    return False, 'Motion never occurred! Stop activated'
            yield dsleep(wait_for_motion)
            mdata = self.data['status']['summary']
        moving = True
        while moving:
            acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                          'block_name': 'ACU_upload',
                          'data': self.data['uploads']
                          }
            self.agent.publish_to_feed('acu_upload', acu_upload)
            mdata = self.data['status']['summary']
            ve = round(mdata['Elevation_current_velocity'], 2)
            va = round(mdata['Azimuth_current_velocity'], 2)
            if (ve != 0.0) or (va != 0.0):
                moving = True
                yield dsleep(wait_for_motion)
            else:
                moving = False
                mdata = self.data['status']['summary']
                pe = round(mdata['Elevation_current_position'], round_int)
                pa = round(mdata['Azimuth_current_position'], round_int)
                if pe != el or pa != az:
                    yield self.acu_control.stop()
                    self.log.warn('Stopped before reaching commanded point!')
                    return False, 'Something went wrong!'
                modes = (mdata['Azimuth_mode'], mdata['Elevation_mode'])
                if modes != ('Preset', 'Preset'):
                    return False, 'Fault triggered!'
        if end_stop:
            yield self.acu_control.stop()
            self.log.info('Stop mode activated')
        self.data['uploads']['Start_Azimuth'] = 0.0
        self.data['uploads']['Start_Elevation'] = 0.0
        self.data['uploads']['Command_Type'] = 0
        self.data['uploads']['Preset_Azimuth'] = 0.0
        self.data['uploads']['Preset_Elevation'] = 0.0
        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload, from_reactor=True)
        self._set_job_done('control')
        return True, 'Pointing completed'

    @inlineCallbacks
    def set_boresight(self, session, params):
        """set_boresight(b=None, end_stop=False)

        **Task** - Move the telescope to a particular third-axis angle.

        Parameters:
            b (float): destination angle for boresight rotation
            end_stop (bool): put axes in Stop mode after motion

        """
        ok, msg = self._try_set_job('control')
        if not ok:
            return ok, msg
        bs_destination = params.get('b')
        # yield self.acu_control.stop()
        yield dsleep(5)
        self.data['uploads']['Start_Boresight'] = self.data['status']['summary']['Boresight_current_position']
        self.data['uploads']['Command_Type'] = 1
        self.data['uploads']['Preset_Boresight'] = bs_destination
        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload)
        yield self.acu_control.go_3rd_axis(bs_destination)
        current_position = self.data['status']['summary']['Boresight_current_position']
        while current_position != bs_destination:
            yield dsleep(1)
            acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                          'block_name': 'ACU_upload',
                          'data': self.data['uploads']
                          }
            self.agent.publish_to_feed('acu_upload', acu_upload)
            current_position = self.data['status']['summary']['Boresight_current_position']
        if params.get('end_stop'):
            yield self.acu_control.stop()
        self.data['uploads']['Start_Boresight'] = 0.0
        self.data['uploads']['Command_Type'] = 0
        self.data['uploads']['Preset_Boresight'] = 0.0
        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload)
        self._set_job_done('control')
        return True, 'Moved to new 3rd axis position'

    @inlineCallbacks
    def stop_and_clear(self, session, params):
        """stop_and_clear()

        **Task** - Change the azimuth and elevation modes to Stop and clear
        points uploaded to the stack.

        """
        ok, msg = self._try_set_job('control')
        if not ok:
            self._set_job_done('control')
            yield dsleep(0.1)
            self._try_set_job('control')
        self.log.info('_try_set_job ok')
        # yield self.acu.stop()
        yield self.acu_control.mode('Stop')
        self.log.info('Stop called')
        yield dsleep(5)
        yield self.acu_control.http.Command('DataSets.CmdTimePositionTransfer',
                                            'Clear Stack')
        yield dsleep(0.1)
        self.log.info('Cleared stack.')
        self._set_job_done('control')
        return True, 'Job completed'

    @inlineCallbacks
    def fromfile_scan(self, session, params=None):
        """fromfile_scan(filename=None, simulator=None)

        **Task** - Upload and execute a scan pattern from numpy file.

        Parameters:
            filename (str): full path to desired numpy file. File contains
                an array of three lists ([list(times), list(azimuths),
                list(elevations)]).  Times begin from 0.0.
            simulator (bool): toggle for using the ACU simulator.
        """
        filename = params.get('filename')
        simulator = params.get('simulator')
        times, azs, els, vas, ves, azflags, elflags = sh.from_file(filename)
        if min(azs) <= self.motion_limits['azimuth']['lower'] or max(azs) >= self.motion_limits['azimith']['upper']:
            return False, 'Azimuth location out of range!'
        if min(els) <= self.motion_limits['elevation']['lower'] or max(els) >= self.motion_limits['elevation']['upper']:
            return False, 'Elevation location out of range!'
        yield self._run_specified_scan(session, times, azs, els, vas, ves, azflags, elflags, azonly=False, simulator=simulator)
        yield True, 'Track completed'

    @ocs_agent.param('azpts', type=tuple)
    @ocs_agent.param('el', type=float)
    @ocs_agent.param('azvel', type=float)
    @ocs_agent.param('acc', type=float)
    @ocs_agent.param('ntimes', type=int)
    @ocs_agent.param('azonly', type=bool)
    @ocs_agent.param('simulator', default=False, type=bool)
    @inlineCallbacks
    def constant_velocity_scan(self, session, params=None):
        """constant_velocity_scan(azpts=None, el=None, azvel=None, acc=None, \
                                  ntimes=None, azonly=None, simulator=False)

        **Task** - Run a constant velocity scan.

        Parameters:
            azpts (tuple): spatial endpoints of the azimuth scan
            el (float): elevation (constant) throughout the scan
            azvel (float): velocity of the azimuth axis
            acc (float): acceleration of the turnaround
            ntimes (int): number of times the platform traverses
                between azimuth endpoints
            azonly (bool): option for scans with azimuth motion and
                elevation in Stop
            simulator (bool): toggle option for ACU simulator

        """
        azpts = params['azpts']
        el = params['el']
        azvel = params['azvel']
        acc = params['acc']
        ntimes = params['ntimes']
        azonly = params['azonly']
        simulator = params['simulator']
        if abs(acc) > self.motion_limits['acc']:
            raise ocs_agent.ParamError('Acceleration too great!')
        if min(azpts) <= self.motion_limits['azimuth']['lower'] or max(azpts) >= self.motion_limits['azimuth']['upper']:
            raise ocs_agent.ParamError('Azimuth location out of range!')
        if el <= self.motion_limits['elevation']['lower'] or el >= self.motion_limits['elevation']['upper']:
            raise ocs_agent.ParamError('Elevation location out of range!')
        times, azs, els, vas, ves, azflags, elflags = sh.constant_velocity_scanpoints(azpts, el, azvel, acc, ntimes)
        yield self._run_specified_scan(session, times, azs, els, vas, ves, azflags, elflags, azonly, simulator)
        return True, 'Track completed.'

    @inlineCallbacks
    def _run_specified_scan(self, session, times, azs, els, vas, ves, azflags, elflags, azonly, simulator):
        ok, msg = self._try_set_job('control')
        if not ok:
            return ok, msg
        self.log.info('_try_set_job ok')

        start_az = azs[0]
        start_el = els[0]
        end_az = azs[-1]
        end_el = els[-1]

        self.data['uploads']['Start_Azimuth'] = start_az
        self.data['uploads']['Start_Elevation'] = start_el
        self.data['uploads']['Command_Type'] = 2

        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload)

        # Follow the scan in ProgramTrack mode, then switch to Stop mode
        all_lines = sh.ptstack_format(times, azs, els, vas, ves, azflags, elflags)
        self.log.info('all_lines generated')
        self.data['uploads']['PtStack_Lines'] = 'True'
        if azonly:
            yield self.acu_control.azmode('ProgramTrack')
        else:
            yield self.acu_control.mode('ProgramTrack')
        m = yield self.acu_control.mode()
        print(m)
        self.log.info('mode is now ProgramTrack')
        if simulator:
            group_size = len(all_lines)
        else:
            group_size = 120
        spec = {'times': times,
                'azs': azs,
                'els': els,
                'vas': vas,
                'ves': ves,
                'azflags': azflags,
                'elflags': elflags,
                }
        while len(all_lines):
            upload_lines = all_lines[:group_size]
            all_lines = all_lines[group_size:]
            upload_vals = front_group(spec, group_size)
            spec = pop_first_vals(spec, group_size)

            for u in range(len(upload_vals['azs'])):
                self.data['uploads']['PtStack_Time'] = upload_lines[u].split(';')[0]
                self.data['uploads']['PtStack_Azimuth'] = upload_vals['azs'][u]
                self.data['uploads']['PtStack_Elevation'] = upload_vals['els'][u]
                self.data['uploads']['PtStack_AzVelocity'] = upload_vals['vas'][u]
                self.data['uploads']['PtStack_ElVelocity'] = upload_vals['ves'][u]
                self.data['uploads']['PtStack_AzFlag'] = upload_vals['azflags'][u]
                self.data['uploads']['PtStack_ElFlag'] = upload_vals['elflags'][u]
                self.data['uploads']['PtStack_ctime'] = uploadtime_to_ctime(self.data['uploads']['PtStack_Time'], int(self.data['status']['summary']['Year']))
                acu_upload = {'timestamp': self.data['uploads']['PtStack_ctime'],
                              'block_name': 'ACU_upload',
                              'data': self.data['uploads']
                              }
                self.agent.publish_to_feed('acu_upload', acu_upload, from_reactor=True)
            text = ''.join(upload_lines)
            free_positions = self.data['status']['summary']['Free_upload_positions']
            while free_positions < 9899:
                free_positions = self.data['status']['summary']['Free_upload_positions']
                yield dsleep(0.1)
            yield self.acu_control.http.UploadPtStack(text)
            self.log.info('Uploaded a group')
        self.log.info('No more lines to upload')
        free_positions = self.data['status']['summary']['Free_upload_positions']
        while free_positions < 9999:
            yield dsleep(0.1)
            modes = (self.data['status']['summary']['Azimuth_mode'],
                     self.data['status']['summary']['Elevation_mode'])
            if azonly:
                if modes[0] != 'ProgramTrack':
                    return False, 'Azimuth mode no longer ProgramTrack!!'
            else:
                if modes != ('ProgramTrack', 'ProgramTrack'):
                    return False, 'Fault triggered (not ProgramTrack)!'
            free_positions = self.data['status']['summary']['Free_upload_positions']
        self.log.info('No more points in the queue')
        current_az = self.data['broadcast']['Corrected_Azimuth']
        current_el = self.data['broadcast']['Corrected_Elevation']
        while round(current_az - end_az, 1) != 0.:
            self.log.info('Waiting to settle at azimuth position')
            yield dsleep(0.1)
            current_az = self.data['broadcast']['Corrected_Azimuth']
        if not azonly:
            while round(current_el - end_el, 1) != 0.:
                self.log.info('Waiting to settle at elevation position')
                yield dsleep(0.1)
                current_el = self.data['broadcast']['Corrected_Elevation']
        yield dsleep(self.sleeptime)
        yield self.acu_control.stop()
        self.data['uploads']['Start_Azimuth'] = 0.0
        self.data['uploads']['Start_Elevation'] = 0.0
        self.data['uploads']['Command_Type'] = 0
        self.data['uploads']['PtStack_Lines'] = 'False'
        self.data['uploads']['PtStack_Time'] = '000, 00:00:00.000000'
        self.data['uploads']['PtStack_Azimuth'] = 0.0
        self.data['uploads']['PtStack_Elevation'] = 0.0
        self.data['uploads']['PtStack_AzVelocity'] = 0.0
        self.data['uploads']['PtStack_ElVelocity'] = 0.0
        self.data['uploads']['PtStack_AzFlag'] = 0
        self.data['uploads']['PtStack_ElFlag'] = 0
        acu_upload = {'timestamp': self.data['status']['summary']['ctime'],
                      'block_name': 'ACU_upload',
                      'data': self.data['uploads']
                      }
        self.agent.publish_to_feed('acu_upload', acu_upload)
        self._set_job_done('control')
        return True

    @inlineCallbacks
    def generate_scan(self, session, params):
        """generate_scan(az_endpoint1=None, az_endpoint2=None, az_speed=None, \
                         acc=None, el_endpoint1=None, el_endpoint2=None, \
                         el_speed=None, num_batches=None, start_time=None, \
                         wait_to_start=None, step_time=None, batch_size=None, \
                         az_start=None)

        **Process** - Scan generator, currently only works for
        constant-velocity az scans with fixed elevation.

        Parameters:
            az_endpoint1 (float): first endpoint of a linear azimuth scan
            az_endpoint2 (float): second endpoint of a linear azimuth scan
            az_speed (float): azimuth speed for constant-velocity scan
            acc (float): turnaround acceleration for a constant-velocity scan
            el_endpoint1 (float): first endpoint of elevation motion
            el_endpoint2 (float): second endpoint of elevation motion. For dev,
                currently both el endpoints should be equal
            el_speed (float): speed of motion for a scan with changing
                elevation. For dev, currently set to 0.0
            num_batches (int or None): sets the number of batches for the
                generator to create. Default value is None (interpreted as infinite
                batches).
            start_time (float or None): a ctime at which to start the scan.
                Default is None, interpreted as now
            wait_to_start (float): number of seconds to wait before starting a
                scan. Default is 3 seconds
            step_time (float): time between points on the constant-velocity
                parts of the motion. Default is 0.1 s. Minimum 0.05 s
            batch_size (int): number of values to produce in each iteration.
                Default is 500. Batch size is reset to the length of one leg of the
                motion if num_batches is not None.
            az_start (str): part of the scan to start at. Options are:
                'az_endpoint1', 'az_endpoint2', 'mid_inc' (start in the middle of
                the scan and start with increasing azimuth), 'mid_dec' (start in
                the middle of the scan and start with decreasing azimuth).
        """
        ok, msg = self._try_set_job('control')
        if not ok:
            return ok, msg
        self.log.info('_try_set_job ok')
        az_endpoint1 = params.get('az_endpoint1')
        az_endpoint2 = params.get('az_endpoint2')
        az_speed = params.get('az_speed')
        acc = params.get('acc')
        el_endpoint1 = params.get('el_endpoint1')
        scan_params = {k: params.get(k) for k in ['num_batches', 'start_time',
                       'wait_to_start', 'step_time', 'batch_size', 'az_start']
                       if params.get(k) is not None}
        el_endpoint2 = params.get('el_endpoint2', el_endpoint1)
        el_speed = params.get('el_speed', 0.0)

        yield self.acu_control.stop()
        g = sh.generate_constant_velocity_scan(az_endpoint1=az_endpoint1,
                                               az_endpoint2=az_endpoint2,
                                               az_speed=az_speed, acc=acc,
                                               el_endpoint1=el_endpoint1,
                                               el_endpoint2=el_endpoint2,
                                               el_speed=el_speed,
                                               **scan_params)
        self.acu_control.mode('ProgramTrack')
        while self.jobs['control'] == 'run':
            lines = next(g)
            current_lines = lines
            group_size = 250
            while len(current_lines):
                upload_lines = current_lines[:group_size]
                text = ''.join(upload_lines)
                current_lines = current_lines[group_size:]
                free_positions = self.data['status']['summary']['Free_upload_positions']
                while free_positions < 5099:
                    yield dsleep(0.1)
                    free_positions = self.data['status']['summary']['Free_upload_positions']
                yield self.acu_control.http.UploadPtStack(text)
        yield self.acu_control.stop()
        self._set_job_done('control')
        return True, 'Track generation ended cleanly'


def add_agent_args(parser_in=None):
    if parser_in is None:
        parser_in = argparse.ArgumentParser()
    pgroup = parser_in.add_argument_group('Agent Options')
    pgroup.add_argument("--acu_config")
    return parser_in


def main(args=None):
    parser = add_agent_args()
    args = site_config.parse_args(agent_class='ACUAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    _ = ACUAgent(agent, args.acu_config)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

ON_RTD = os.environ.get('READTHEDOCS') == 'True'
if not ON_RTD:
    # yes I shouldn't have named that module agent
    from xy_agent.xy_connect import XY_Stage


class LATRtXYStageAgent:
    """
    Agent for connecting to the LATRt XY Stages.

    Args:
        ip_addr: IP address where RPi server is running.
        port: Port the RPi Server is listening on.
        mode: 'acq': Start data acquisition on initialize.
        samp: Default sampling frequency in Hz.
    """

    def __init__(self, agent, ip_addr, port, mode=None, samp=2):

        self.ip_addr = ip_addr
        self.port = int(port)

        self.xy_stage = None
        self.initialized = False
        self.take_data = False
        self.is_moving = False

        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        if mode == 'acq':
            self.auto_acq = True
        else:
            self.auto_acq = False
        self.sampling_frequency = float(samp)

        # register the position feeds
        agg_params = {
            'frame_length': 10 * 60,  # [sec]
        }

        self.agent.register_feed('positions',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('_')
    def init_xy_stage(self, session, params=None):
        """init_xy_stage()

        **Task** - Perform first time setup for communication with XY stages.

        """

        self.log.debug("Trying to acquire lock")
        with self.lock.acquire_timeout(timeout=0, job='init') as acquired:
            # Locking mechanism stops code from proceeding if no lock acquired
            if not acquired:
                self.log.warn("Could not start init because {} is already running".format(self.lock.job))
                return False, "Could not acquire lock."
            # Run the function you want to run
            self.log.debug("Lock Acquired Connecting to Stages")

            self.xy_stage = XY_Stage(self.ip_addr, self.port)
            self.xy_stage.init_stages()
            print("XY Stages Initialized")

        # This part is for the record and to allow future calls to proceed,
        # so does not require the lock
        self.initialized = True
        if self.auto_acq:
            self.agent.start('acq')
        return True, 'XY Stages Initialized.'

    @ocs_agent.param('distance', type=float)
    @ocs_agent.param('velocity', type=float, check=lambda x: 0 <= x < 1.2)
    def move_x_cm(self, session, params):
        """move_x_cm(distance, velocity)

        **Task** - Move the X axis.

        Parameters:
            distance (float): Distance to move in cm.
            velocity (float): Velocity to move at. Must be less than 1.2.

        """

        with self.lock.acquire_timeout(timeout=3, job='move_x_cm') as acquired:
            if not acquired:
                self.log.warn(f"Could not start x move because lock held by {self.lock.job}")
                return False
            self.xy_stage.move_x_cm(params.get('distance', 0), params.get('velocity', 1))

        time.sleep(1)
        while True:
            # data acquisition updates the moving field if it is running
            if not self.take_data:
                with self.lock.acquire_timeout(timeout=3, job='move_x_cm') as acquired:
                    if not acquired:
                        self.log.warn(f"Could not check because lock held by {self.lock.job}")
                        return False, "Could not acquire lock"
                    self.is_moving = self.xy_stage.moving

            if not self.is_moving:
                break
        return True, "X Move Complete"

    @ocs_agent.param('distance', type=float)
    @ocs_agent.param('velocity', type=float, check=lambda x: 0 <= x < 1.2)
    def move_y_cm(self, session, params):
        """move_y_cm(distance, velocity)

        **Task** - Move the Y axis.

        Parameters:
            distance (float): Distance to move in cm.
            velocity (float): Velocity to move at. Must be less than 1.2.

        """

        with self.lock.acquire_timeout(timeout=3, job='move_y_cm') as acquired:
            if not acquired:
                self.log.warn(f"Could not start y move because lock held by {self.lock.job}")
                return False, "could not acquire lock"
            self.xy_stage.move_y_cm(params.get('distance', 0), params.get('velocity', 1))

        time.sleep(1)
        while True:
            # data acquisition updates the moving field if it is running
            if not self.take_data:
                with self.lock.acquire_timeout(timeout=3, job='move_y_cm') as acquired:
                    if not acquired:
                        self.log.warn(f"Could not check for move because lock held by {self.lock.job}")
                        return False, "could not acquire lock"
                    self.is_moving = self.xy_stage.moving
            if not self.is_moving:
                break
        return True, "Y Move Complete"

    @ocs_agent.param('position', type=tuple)
    def set_position(self, session, params):
        """set_position(position)

        **Task** - Set position of the XY stage.

        Parameters:
            position (tuple): (X, Y) position.

        """
        with self.lock.acquire_timeout(timeout=3, job='set_position') as acquired:
            if not acquired:
                self.log.warn(f"Could not set position because lock held by {self.lock.job}")
                return False, "Could not acquire lock"

            self.xy_stage.position = params['position']
        return True, "Position Updated"

    @ocs_agent.param('sampling_frequency', type=float)
    def acq(self, session, params=None):
        """acq(sampling_frequency=2)

        **Process** - Run data acquisition.

        Parameters:
            sampling_frequency (float): Sampling rate to acquire data at.
                Defaults to value set in site config file (or 2 Hz if
                unspecified.)

        Notes:
            The most recent positions are stored in the session.data object in the
            format::

                >>> response.session['data']
                {"positions":
                    {"x": x position in cm,
                     "y": y position in cm}
                }

        """
        if params is None:
            params = {}

        f_sample = params.get('sampling_frequency', self.sampling_frequency)
        pm = Pacemaker(f_sample, quantize=True)

        if not self.initialized or self.xy_stage is None:
            raise Exception("Connection to XY Stages not initialized")

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            self.log.info(f"Starting Data Acquisition for XY Stages at {f_sample} Hz")
            session.set_status('running')
            self.take_data = True
            last_release = time.time()

            while self.take_data:
                if time.time() - last_release > 1.:
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Could not re-acquire lock now held by {self.lock.job}.")
                        return False, "could not re-acquire lock"
                    last_release = time.time()
                pm.sleep()

                data = {'timestamp': time.time(), 'block_name': 'positions', 'data': {}}
                pos = self.xy_stage.position
                self.is_moving = self.xy_stage.moving

                data['data']['x'] = pos[0]
                data['data']['y'] = pos[1]

                self.agent.publish_to_feed('positions', data)
                session.data.update(data['data'])
        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        params:
            dict: {}
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running.'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--port')
    pgroup.add_argument('--mode')
    pgroup.add_argument('--sampling_frequency')
    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class='LATRtXYStageAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    xy_agent = LATRtXYStageAgent(agent, args.ip_address, args.port, args.mode, args.sampling_frequency)

    agent.register_task('init_xy_stage', xy_agent.init_xy_stage)
    agent.register_task('move_x_cm', xy_agent.move_x_cm)
    agent.register_task('move_y_cm', xy_agent.move_y_cm)
    agent.register_task('set_position', xy_agent.set_position)

    agent.register_process('acq', xy_agent.acq, xy_agent._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()
